/*	License
 * 		inside EU: EUPL
 * 		outside EU: GPL or EUPL
 *  The program comes AS IS and without any warranty unless required by law
 *  (C) 2012 thecodemonkey
 */

package scheduler;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;



public class Scheduler2 implements IScheduler {
	protected List<SchedulerRunnable2> queued;
	protected List<SchedulerRunnable2> stopped;
	protected Map<Integer, SchedulerRunnable2> active;
	
	protected Object lock = new Object();
	
	protected boolean running;
	protected int maxThreads, threadCounter;//, taskCounter, numberOfTasks;
	protected SchedulerClient client;
	
	public Scheduler2(SchedulerClient client, int maxThreads) {
		this.client=client;
		queued=Collections.synchronizedList(new LinkedList<SchedulerRunnable2>());
		stopped=Collections.synchronizedList(new LinkedList<SchedulerRunnable2>());
		active=Collections.synchronizedMap(new HashMap<Integer,SchedulerRunnable2>());
		this.maxThreads=maxThreads;
	}
	
	public void setMaxThreads(int maxThreads) {
		this.maxThreads = maxThreads;
	}
	
	@Override
	public void start() {
		synchronized (lock) {
			running=true;
			System.out.println("start");
			int k=Math.min(queued.size(), maxThreads);
			
			for (int i=0;i<k;i++) {
				startTask();
			}
		}
	}

	@Override
	public int addTaskStopped(SchedulerRunnable2 t) {
		synchronized(lock){
			stopped.add(t);
		}
		return t.getId();
	}
	 
	@Override
	public int addTask(SchedulerRunnable2 t) {
		synchronized(lock) {
			queued.add(t);
			if (running) {
				if (threadCounter<maxThreads) {
					startTask();
				}
			}
		}
		return t.getId();
	}
	
	@Override
	public void stopAll() {
		synchronized (lock) {
			running=false;
			Iterator<SchedulerRunnable2>it =active.values().iterator();
			while (it.hasNext()) {
				SchedulerRunnable2 t=it.next();
				t.stop();
			}
			
			stopped.addAll(active.values());				
			active.clear();
			stopped.addAll(queued);				
			queued.clear();
		}
	}
	
	@Override
	public SchedulerRunnable2 remove(int id) {
		synchronized (lock) {
			SchedulerRunnable2 t=active.get(id);
			
			if (t!=null) {
				t.stop();
				SchedulerRunnable2 thread = active.remove(id);
				
				if (!queued.isEmpty()) {
					startTask();
				} else if (active.isEmpty() &&stopped.isEmpty()) {
					running = false;
					client.allDone();
				}
						
				return thread;
			} else {
				
				for (SchedulerRunnable2 thread:queued) {
					if (thread.id==id) {
						queued.remove(thread);
						return thread;
					}
				}
				
				for (SchedulerRunnable2 thread:stopped) {
					if (thread.id==id) {
						stopped.remove(thread);
						return thread;
					}
				}
				
				return null;
			}
		}
		
	}
	
	@Override
	public void removeAll() {
		stopAll();
		queued.clear();
		stopped.clear();
	}
	
	@Override
	public SchedulerRunnable2 stop(int id) {
		synchronized (lock){
			SchedulerRunnable2 t=active.remove(id);
			
			if (t!=null) {
				t.stop();
				stopped.add(t);
				if (active.isEmpty() && queued.isEmpty()) running=false;
			} else {
				
				for (SchedulerRunnable2 thread:queued) {
					if (thread.id==id) {
						queued.remove(thread);
						synchronized (stopped) {
							stopped.add(thread);
						}
						return thread;
					}
				}
				
			}
			return t;
		}
	}
	
	protected void startTask() {
		if (running) {
			SchedulerRunnable2 thread=queued.remove(0);
			active.put(thread.id,  thread);
			threadCounter++; 
			new SchedulerThread(thread).start();
		}
	}
	
	protected class SchedulerThread extends Thread {
		
		private SchedulerRunnable2 concreteTask;
		
		public SchedulerThread(SchedulerRunnable2 concreteTask) {
			this.concreteTask=concreteTask;
		}
		
		public SchedulerRunnable2 getConcreteTask() {
			return concreteTask;
		}
		
		@Override
		public void run() {
			concreteTask.run();
			synchronized (lock) {
				threadCounter--;
				active.remove(concreteTask.getId());
				
				if (concreteTask.errorOccoured()){				
					concreteTask.stop();
					stopped.add(concreteTask);
				}
				if (!queued.isEmpty()) {
					startTask();
				} else if (active.isEmpty()) {
					running = false;
					
					if (stopped.isEmpty()) {
						client.allDone();
					}
				}
			}
			
		}
	}
	
	@Override
	public boolean isRunning() {
		synchronized (lock) {
			return running;
		}
	}
	

	@Override
	public Collection<SchedulerRunnable2> getActive() {
		synchronized (lock) {
			return new ArrayList<SchedulerRunnable2>(active.values());
			
		}
		
	}
	
	@Override
	public Collection<SchedulerRunnable2> getUnfinished() {
		synchronized (lock) {
			List<SchedulerRunnable2> list=new ArrayList<>(active.size()+queued.size()+stopped.size());
			list.addAll(active.values());
			list.addAll(queued);
			list.addAll(stopped);
			return list;
		}
	}

	/**
	 * Is active or queued
	 */
	@Override
	public boolean isRunning(int id) {
		synchronized (lock) {
			if (active.containsKey(id)) {
				return true;
			}
		
			for (SchedulerRunnable2 sr:queued) {
				if (sr.getId()==id) return true;
			}
		}
		return false;
	}
	
	@Override
	public SchedulerRunnable2 enqueueStopped(int id) {
		synchronized (lock) {
			Iterator<SchedulerRunnable2> it=stopped.iterator();
			SchedulerRunnable2 sr = null;
			while (it.hasNext()) {
				SchedulerRunnable2 task=it.next();
				if (task.getId()==id) {
					task.enqueue();
					
					queued.add(0,task);
					
					it.remove();
					sr=task;
					break;
				}
			}
			return sr;
		}
	}
	
	@Override
	public void enqueueAllStopped() {
		synchronized (lock) {
			for (SchedulerRunnable2 thread:stopped)
				thread.enqueue();
			queued.addAll(stopped);
			stopped.clear();
		}
	}
	
	@Override
	public SchedulerRunnable2 getStoppedById(int id) {
		synchronized (lock) {
			for (SchedulerRunnable2 r:stopped) {
				if (r.id == id) return r;
			}
		}
		return null;
	}
	
	@Override
	public Collection<SchedulerRunnable2> getStopped() {
		synchronized (lock) {
			return new ArrayList<>(stopped);
		}
	}	

}
